/*
 * Copyright (C) 2007 Brice Figureau
 * shm_zone and rbtree code Copyright (c) 2002-2007 Igor Sysoev
 */
/*
 * Copyright (C) 2009   Lev Alexandrovsky <lev0666@yandex.ru>, 
                        Alexey Guseynov <aguseynov@gmail.com>
 * integration with Nginx upload module (v 2.0.10)
 */

#include "progress.h"

static ngx_str_t x_progress_id = ngx_string("X-Progress-ID");

static ngx_int_t ngx_http_uploadprogress_init_zone(ngx_shm_zone_t * shm_zone, void *data);
static void ngx_clean_old_connections(ngx_event_t * ev); 
static  void ngx_http_uploadprogress_cleanup(void *data);

static ngx_str_t*
get_tracking_id(ngx_http_request_t * r)
{
    u_char                          *p, *start_p;
    ngx_uint_t                       i;
    ngx_list_part_t                 *part;
    ngx_table_elt_t                 *header;
    ngx_str_t                       *ret;
	
    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, "upload-progress: get_tracking_id");

    part = &r->headers_in.headers.part;
    header = part->elts;

    for (i = 0; /* void */ ; i++) {

        if (i >= part->nelts) {
            if (part->next == NULL) {
                break;
            }

            part = part->next;
            header = part->elts;
            i = 0;
        }

        if (header[i].key.len == x_progress_id.len
            && ngx_strncasecmp(header[i].key.data, x_progress_id.data,
                           header[i].key.len) == 0) {
            ret = ngx_calloc(sizeof(ngx_str_t), r->connection->log );
            ret->data = header[i].value.data;
            ret->len = header[i].value.len;
            ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, 
                            "upload-progress: get_tracking_id found header: %V", ret);
            return ret;
        }
    }

    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, 
                    "upload-progress: get_tracking_id no header found");

    /* not found, check as a request arg */
    if (r->args.len) {
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, 
                       "upload-progress: get_tracking_id no header found but args present");
        i = 0;
        p = r->args.data;
        do {
            ngx_uint_t len = r->args.len - (p - r->args.data);
            if (len >= 14 && ngx_strncasecmp(p, (u_char*)"X-Progress-ID=", 14) == 0) {
              ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, 
                             "upload-progress: get_tracking_id found args: %s",p);
                i = 1;
                break;
            }
            if (len<=0)
                break;
        } 
        while(p++);

        if (i) {
            start_p = p += 14;
            while (p < r->args.data + r->args.len) {
                if (*p++ != '&') {
                    continue;
                }
            }

            ret = ngx_calloc(sizeof(ngx_str_t), r->connection->log);
            ret->data = start_p;
            ret->len = p - start_p;
            ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, 
                           "upload-progress: get_tracking_id found args: %V",ret);
            return ret;
        }
    }

    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, 
                   "upload-progress: get_tracking_id no id found");
    return NULL;
}



static ngx_http_uploadprogress_node_t *
find_node(ngx_str_t * id, ngx_http_uploadprogress_ctx_t * ctx, ngx_log_t * log)
{
    uint32_t                         hash;
    ngx_rbtree_node_t               *node, *sentinel;
    ngx_int_t                        rc;
    ngx_http_uploadprogress_node_t  *up;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0, "upload-progress: find_node %V", id);

    hash = ngx_crc32_short(id->data, id->len);

    node = ctx->rbtree->root;
    sentinel = ctx->rbtree->sentinel;

    while (node != sentinel) {

        if (hash < node->key) {
            node = node->left;
            continue;
        }

        if (hash > node->key) {
            node = node->right;
            continue;
        }

        /* hash == node->key */

        do {
            up = (ngx_http_uploadprogress_node_t *) node;

            rc = ngx_memn2cmp(id->data, up->data, id->len, (size_t) up->len);

            if (rc == 0) {
                ngx_log_debug0(NGX_LOG_DEBUG_HTTP, log, 0,
                               "upload-progress: found node");
                return up;
            }

            node = (rc < 0) ? node->left : node->right;

        } while (node != sentinel && hash == node->key);

        break;
    }
    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, log, 0, "upload-progress: can't find node");
    return NULL;
}
 

extern char* upload_progress_handler(ngx_conf_t * cf, ngx_module_t *ngx_http_upload_module)
{
    ssize_t                          n;
    ngx_str_t                       *value;
    ngx_shm_zone_t                  *shm_zone;
    ngx_http_uploadprogress_ctx_t   *ctx;

    value = cf->args->elts;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, cf->log, 0,
                   "ngx_upload_progress name: %V", &value[1]);

    ctx = ngx_pcalloc(cf->pool, sizeof(ngx_http_uploadprogress_ctx_t));
    if (ctx == NULL) {
        return NGX_CONF_ERROR;
    }

    ctx->list_head.prev = NULL;
    ctx->list_head.next = &ctx->list_tail;

    ctx->list_tail.prev = &ctx->list_head;
    ctx->list_tail.next = NULL;

    n = ngx_parse_size(&value[2]);

    if (n == NGX_ERROR) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "invalid size of track_uploads \"%V\"", &value[2]);
        return NGX_CONF_ERROR;
    }

    if (n < (ngx_int_t) (8 * ngx_pagesize)) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "track_uploads \"%V\" is too small", &value[1]);
        return NGX_CONF_ERROR;
    }

    shm_zone = ngx_shared_memory_add(cf, &value[1], n,
                                     ngx_http_upload_module);
    if (shm_zone == NULL) {
        return NGX_CONF_ERROR;
    }

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, cf->log, 0,
                   "ngx_upload_progress name: %V, szhm_zone: %p", &value[1],
                   shm_zone);

    if (shm_zone->data) {
        ctx = shm_zone->data;

        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "track_uploads \"%V\" is already created", &value[1]);
        return NGX_CONF_ERROR;
    }

    shm_zone->init = ngx_http_uploadprogress_init_zone;
    shm_zone->data = ctx;

    return NGX_CONF_OK;
}

static void
ngx_http_uploadprogress_rbtree_insert_value(ngx_rbtree_node_t * temp,
                                            ngx_rbtree_node_t * node,
                                            ngx_rbtree_node_t * sentinel)
{
    ngx_http_uploadprogress_node_t  *upn, *upnt;

    for (;;) {

        if (node->key < temp->key) {

            if (temp->left == sentinel) {
                temp->left = node;
                break;
            }

            temp = temp->left;

        } else if (node->key > temp->key) {

            if (temp->right == sentinel) {
                temp->right = node;
                break;
            }

            temp = temp->right;

        } else {                /* node->key == temp->key */

            upn = (ngx_http_uploadprogress_node_t *) node;
            upnt = (ngx_http_uploadprogress_node_t *) temp;

            if (ngx_memn2cmp(upn->data, upnt->data, upn->len, upnt->len) < 0) {

                if (temp->left == sentinel) {
                    temp->left = node;
                    break;
                }

                temp = temp->left;

            } else {

                if (temp->right == sentinel) {
                    temp->right = node;
                    break;
                }

                temp = temp->right;
            }
        }
    }

    node->parent = temp;
    node->left = sentinel;
    node->right = sentinel;
    ngx_rbt_red(node);
}

static ngx_int_t
ngx_http_uploadprogress_init_zone(ngx_shm_zone_t * shm_zone, void *data)
{
    ngx_http_uploadprogress_ctx_t   *octx = data;

    ngx_slab_pool_t                 *shpool;
    ngx_rbtree_node_t               *sentinel;
    ngx_http_uploadprogress_ctx_t   *ctx;

    ctx = shm_zone->data;

    if (octx) {
        ctx->rbtree = octx->rbtree;
        return NGX_OK;
    }

    shpool = (ngx_slab_pool_t *) shm_zone->shm.addr;

    ctx->rbtree = ngx_slab_alloc(shpool, sizeof(ngx_rbtree_t));
    if (ctx->rbtree == NULL) {
        return NGX_ERROR;
    }

    sentinel = ngx_slab_alloc(shpool, sizeof(ngx_rbtree_node_t));
    if (sentinel == NULL) {
        return NGX_ERROR;
    }

    ngx_rbtree_sentinel_init(sentinel);

    ctx->rbtree->root = sentinel;
    ctx->rbtree->sentinel = sentinel;
    ctx->rbtree->insert = ngx_http_uploadprogress_rbtree_insert_value;

    return NGX_OK;
}

extern char*
track_upload_handler(ngx_http_uploadprogress_conf_t *track_info, u_char *track, ngx_conf_t * cf, ngx_module_t *ngx_http_upload_module)
{
    ngx_str_t *value;

    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, cf->log, 0, "ngx_track_uploads in");

    value = cf->args->elts;

    
	ngx_log_error(NGX_LOG_INFO, cf->log, 0,
                   "ngx_track_uploads name: %V", &value[1]);

    track_info->shm_zone = ngx_shared_memory_add(cf, &value[1], 0,
                                           ngx_http_upload_module);
    if (track_info->shm_zone == NULL) {
        return NGX_CONF_ERROR;
    }

    *track = (u_char) 1;

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, cf->log, 0,
                   "ngx_track_uploads name: %V,szhm_zone: %p", &value[1],
                   track_info->shm_zone);


    track_info->timeout = ngx_parse_time(&value[2], 1);
    if (track_info->timeout == NGX_ERROR) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "track_uploads \"%V\" timeout value invalid", &value[1]);
        return NGX_CONF_ERROR;
    }

    if (track_info->timeout == NGX_PARSE_LARGE_TIME) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "track_uploads \"%V\" timeout value must be less than 68 years", &value[1]);
        return NGX_CONF_ERROR;
    }
  
    return NGX_CONF_OK;
}


extern ngx_int_t
ngx_http_upload_progress_init(ngx_http_request_t * r, ngx_http_uploadprogress_conf_t *track_info, u_char track)
{
    size_t                           n;
    ngx_str_t                       *id;
    uint32_t                         hash;
    ngx_slab_pool_t                 *shpool;
    ngx_rbtree_node_t               *node;
    ngx_http_uploadprogress_ctx_t   *ctx;
    ngx_http_uploadprogress_node_t  *up;
    ngx_http_uploadprogress_cleanup_t *upcln; 
    ngx_pool_cleanup_t              *cln;
		
	
	
    id = get_tracking_id(r);
    if (id == NULL) {
        ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                       "trackuploads no id found in POST upload req");
		
        return NGX_DECLINED;
    }
	if (!track) {
        ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                       "trackuploads not tracking in this location for id: %V", id);
		ngx_free(id);
        return NGX_DECLINED;
    }

    ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                   "trackuploads id found: %V", id);

    

    if (track_info->shm_zone == NULL) {
        ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                       "trackuploads no shm_zone for id: %V", id);
        ngx_free(id);
        return NGX_DECLINED;
    }

    ctx = track_info->shm_zone->data;

    hash = ngx_crc32_short(id->data, id->len);

    ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                   "trackuploads hash %08XD for id: %V", hash, id);

    shpool = (ngx_slab_pool_t *) track_info->shm_zone->shm.addr;

    ngx_shmtx_lock(&shpool->mutex);
	ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                   "shm_zone locked");

    if (find_node(id, ctx, r->connection->log) != NULL) {
        ngx_shmtx_unlock(&shpool->mutex);
        /* already found a node with matching progress ID */
        ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                      "upload_progress: tracking already registered id: %V", id);

        ngx_free(id);
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
	
    cln = ngx_pool_cleanup_add(r->pool, sizeof(ngx_http_uploadprogress_cleanup_t));
    if (cln == NULL) {
        ngx_shmtx_unlock(&shpool->mutex);
        ngx_free(id);
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }
	
    n = sizeof(ngx_http_uploadprogress_node_t)
        + id->len;

    node = ngx_slab_alloc_locked(shpool, n);
    if (node == NULL) {
        ngx_shmtx_unlock(&shpool->mutex);
        ngx_free(id);
        return NGX_HTTP_SERVICE_UNAVAILABLE;
    }
		
    up = (ngx_http_uploadprogress_node_t *) node;

    node->key = hash;
    up->len = (u_char) id->len;
    up->err_status = r->err_status;
	
    up->done = 0;
//    up->rest = r->request_body->rest;
	up->rest = 0;
    up->length = r->headers_in.content_length_n;
    up->timeout = 0;
	
    up->next = ctx->list_head.next;
    up->next->prev = up;
    up->prev = &ctx->list_head;
    ctx->list_head.next = up;
	
    ngx_memcpy(up->data, id->data, id->len);

	ngx_rbtree_insert(ctx->rbtree, node);

    				   
    if (!track_info->cleanup.timer_set) {
        track_info->cleanup.data = track_info->shm_zone;
        track_info->cleanup.handler = ngx_clean_old_connections;
        track_info->cleanup.log = track_info->shm_zone->shm.log;
        ngx_add_timer(&track_info->cleanup, TIMER_FREQUENCY);
    }

    ngx_shmtx_unlock(&shpool->mutex);
	ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                   "shm_zone unlocked");

    cln->handler = ngx_http_uploadprogress_cleanup;
    upcln = cln->data;

    upcln->shm_zone = track_info->shm_zone;
    upcln->node = node;
	
    upcln->timeout = track_info->timeout;
	ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                   "timeout: %T", upcln->timeout);
    upcln->r = r;

    ngx_free(id);

	// ctx = ngx_pcalloc(r->pool, sizeof(ngx_http_upload_ctx_t));
	// if (ctx == NULL) {
     // return NGX_ERROR;
	// }

   // ngx_http_set_ctx(r, ctx, ngx_http_upload_module);
     
    /* finally says to the core we don't handle anything */
    return NGX_OK;
}

static void
ngx_clean_old_connections(ngx_event_t * ev)
{
    ngx_shm_zone_t                  *shm_zone;
    ngx_http_uploadprogress_ctx_t   *ctx;
    ngx_slab_pool_t                 *shpool;
    ngx_rbtree_node_t               *node;
    ngx_http_uploadprogress_node_t  *up, *upprev;
    time_t                           now = ngx_time();
    int                              count = 0;


    /* scan the rbtree */
    shm_zone = ev->data;
    ctx = shm_zone->data;
    shpool = (ngx_slab_pool_t *) shm_zone->shm.addr;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, shm_zone->shm.log, 0,
                   "uploadprogress clean old connections at %T", now);

    ngx_shmtx_lock(&shpool->mutex);
    node = (ngx_rbtree_node_t *) ctx->list_tail.prev;
    for (;;) {


        if (node == &ctx->list_head.node) {
            break;
        }

        up = (ngx_http_uploadprogress_node_t *) node;
        upprev = up->prev;
		
		ngx_log_debug3(NGX_LOG_DEBUG_HTTP, shm_zone->shm.log, 0,		
                       "uploadprogress clean: scanning %08XD (req done %ui) timeout at %T",
                       node->key, up->done, up->timeout);

        if ( (up->done && up->timeout < now) || (ngx_quit || ngx_terminate || ngx_exiting) ) {
            up->next->prev = up->prev;
            up->prev->next = up->next;
			
			ngx_log_debug3(NGX_LOG_DEBUG_HTTP, shm_zone->shm.log, 0,
                           "uploadprogress clean: removing %08XD (req %ui) ",
                           node->key, up->done, up->timeout);

            ngx_rbtree_delete(ctx->rbtree, node);
            ngx_slab_free_locked(shpool, node);
        }
        else {
            ngx_log_debug3(NGX_LOG_DEBUG_HTTP, shm_zone->shm.log, 0,		
                       "uploadprogress clean: found %08XD (req done %ui) timeout at %T",
                       node->key, up->done, up->timeout);
			count++;
			
		}
        node = (ngx_rbtree_node_t *) upprev;
    }


    /* don't reschedule timer if ngx_quit or ngx_terminate && nodes emtpy */
//	if ( count > 0 || !(ngx_quit || ngx_terminate || ngx_exiting)) {
	if ( count > 0 ) {
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, shm_zone->shm.log, 0,
                   "uploadprogress clean old connections restarting timer");
        ngx_add_timer(ev, TIMER_FREQUENCY);       /* trigger again in 60s */
    } else if (ngx_quit || ngx_terminate || ngx_exiting) {
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, shm_zone->shm.log, 0,
                   "uploadprogress clean old connections quitting , no more active connections: not restarting timer");
    }
    ngx_shmtx_unlock(&shpool->mutex);
}



/*
removes the expired node from the upload rbtree
*/
extern void
ngx_http_uploadprogress_cleanup(void *data)
{
    ngx_http_uploadprogress_cleanup_t *upcln = data;
    ngx_slab_pool_t                 *shpool;
    ngx_rbtree_node_t               *node;
    ngx_http_uploadprogress_ctx_t   *ctx; 	
    ngx_http_uploadprogress_node_t  *up;
    ngx_http_request_t              *r;

    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, upcln->shm_zone->shm.log, 0,
                   "uploadprogress cleanup called");
	
    ctx = upcln->shm_zone->data;
    shpool = (ngx_slab_pool_t *) upcln->shm_zone->shm.addr;
    node = upcln->node;
    r = upcln->r;
    up = (ngx_http_uploadprogress_node_t *) node;

    ngx_shmtx_lock(&shpool->mutex);
    
    up->done = 1;               /* mark the original request as done */
    up->timeout = ngx_time() + upcln->timeout;      /* keep tracking for 60s */
    
    if (r != NULL ) {
        ngx_uint_t rc = r->err_status ? r->err_status : r->headers_out.status;
        if (rc >= NGX_HTTP_SPECIAL_RESPONSE) {
            up->err_status = rc;
        }
    }
    
    ngx_shmtx_unlock(&shpool->mutex);

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, upcln->shm_zone->shm.log, 0,
                   "uploadprogress cleanup: connection %08XD to be deleted at %T",
                   node->key, up->timeout);

}

extern char 
*report_upload_handler(ngx_http_uploadprogress_conf_t *track_info, u_char *track, ngx_conf_t * cf, ngx_module_t *ngx_http_upload_module)
{
    ngx_str_t                       *value;

    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, cf->log, 0, "ngx_report_uploads in");

    value = cf->args->elts;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, cf->log, 0,
                   "ngx_report_uploads name: %V", &value[1]);

    track_info->shm_zone = ngx_shared_memory_add(cf, &value[1], 0,
                                           ngx_http_upload_module);
    if (track_info->shm_zone == NULL) {
        return NGX_CONF_ERROR;
    }

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, cf->log, 0,
                   "ngx_report_uploads name: %V, szhm_zone: %p", &value[1],
                   track_info->shm_zone);
	return NGX_CONF_OK;
  
}

extern ngx_int_t
report_handler(ngx_http_request_t * r, ngx_http_uploadprogress_conf_t *track_info)
{
    ngx_str_t                       *id;
    ngx_buf_t                       *b;
    ngx_chain_t                      out;
    ngx_int_t                        rc, size, found=0, done=0, err_status=0;
    off_t                            rest=0, length=0;
    ngx_uint_t                       len, i;
    ngx_slab_pool_t                 *shpool;
        
	ngx_http_uploadprogress_ctx_t   		*ctx;
    ngx_http_uploadprogress_node_t  *up;
    ngx_table_elt_t                 *expires, *cc, **ccp;

    if (r->method != NGX_HTTP_GET && r->method != NGX_HTTP_HEAD) {
        return NGX_HTTP_NOT_ALLOWED;
    }

    rc = ngx_http_discard_request_body(r);

    if (rc != NGX_OK) {
        return rc;
    }

    /* get the tracking id if any */
    id = get_tracking_id(r);


    if (id == NULL) {
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "reportuploads handler cant find id");
        return NGX_DECLINED;
    }

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                   "reportuploads handler found id: %V", id);

    
    if (track_info->shm_zone == NULL) {
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "reportuploads no shm_zone for id: %V", id);
        ngx_free(id);
        return NGX_DECLINED;
    }

    ctx = track_info->shm_zone->data;

    /* get the original connection of the upload */
    shpool = (ngx_slab_pool_t *) track_info->shm_zone->shm.addr;

    ngx_shmtx_lock(&shpool->mutex);

    up = find_node(id, ctx, r->connection->log);
    if (up != NULL) {
        ngx_log_debug5(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "reportuploads found node: %V (rest: %uO, length: %uO, done: %ui, err_status: %ui)", id, up->rest, up->length, up->done, up->err_status);
        rest = up->rest;
        length = up->length;
        done = up->done;
        err_status = up->err_status;
        found = 1;
    } else {
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "reportuploads not found: %V", id);
    }
    ngx_shmtx_unlock(&shpool->mutex);
	ngx_free(id);

    /* send the output */
    r->headers_out.content_type.len = sizeof("text/javascript") - 1;
    r->headers_out.content_type.data = (u_char *) "text/javascript";

    /* force no-cache */
    expires = r->headers_out.expires;

    if (expires == NULL) {

        expires = ngx_list_push(&r->headers_out.headers);
        if (expires == NULL) {
            return NGX_HTTP_INTERNAL_SERVER_ERROR;
        }

        r->headers_out.expires = expires;

        expires->hash = 1;
        expires->key.len = sizeof("Expires") - 1;
        expires->key.data = (u_char *) "Expires";
    }

    len = sizeof("Mon, 28 Sep 1970 06:00:00 GMT");
    expires->value.len = len - 1;

    ccp = r->headers_out.cache_control.elts;
    if (ccp == NULL) {

        if (ngx_array_init(&r->headers_out.cache_control, r->pool,
                           1, sizeof(ngx_table_elt_t *))
            != NGX_OK) {
            return NGX_HTTP_INTERNAL_SERVER_ERROR;
        }

        ccp = ngx_array_push(&r->headers_out.cache_control);
        if (ccp == NULL) {
            return NGX_HTTP_INTERNAL_SERVER_ERROR;
        }

        cc = ngx_list_push(&r->headers_out.headers);
        if (cc == NULL) {
            return NGX_HTTP_INTERNAL_SERVER_ERROR;
        }

        cc->hash = 1;
        cc->key.len = sizeof("Cache-Control") - 1;
        cc->key.data = (u_char *) "Cache-Control";

        *ccp = cc;

    } else {
        for (i = 1; i < r->headers_out.cache_control.nelts; i++) {
            ccp[i]->hash = 0;
        }

        cc = ccp[0];
    }

    expires->value.data = (u_char *) "Thu, 01 Jan 1970 00:00:01 GMT";

    cc->value.len = sizeof("no-cache") - 1;
    cc->value.data = (u_char *) "no-cache";


    if (r->method == NGX_HTTP_HEAD) {
        r->headers_out.status = NGX_HTTP_OK;

        rc = ngx_http_send_header(r);

        if (rc == NGX_ERROR || rc > NGX_OK || r->header_only) {
            return rc;
        }
    }

/*
 There are 4 possibilities
   * request not yet started: found = false
   * request in error:        err_status >= NGX_HTTP_SPECIAL_RESPONSE
   * request finished:        done = true
   * request not yet started but registered:        length==0 && rest ==0
   * reauest in progress:     rest > 0 
 */

    if (!found) {
        size = sizeof("new Object({ 'state' : 'starting' })\r\n");
    } else if (err_status >= NGX_HTTP_SPECIAL_RESPONSE) {
        size = sizeof("new Object({ 'state' : 'error', 'status' : ") + NGX_INT_T_LEN + sizeof(" })\r\n");
    } else if (done) {
        size = sizeof("new Object({ 'state' : 'done' })\r\n");
    } else if ( length == 0 && rest == 0 ) {
        size = sizeof("new Object({ 'state' : 'starting' })\r\n");
    } else {
        size = sizeof("new Object({ 'state' : 'uploading', 'received' : ") + NGX_INT_T_LEN + sizeof(" })\r\n");
        size += sizeof(", 'size' : ") + NGX_INT_T_LEN;
    }

    b = ngx_create_temp_buf(r->pool, size);
    if (b == NULL) {
        return NGX_HTTP_INTERNAL_SERVER_ERROR;
    }

    out.buf = b;
    out.next = NULL;

    if (!found) {
        b->last = ngx_cpymem(b->last, "new Object({ 'state' : 'starting' })\r\n",
                            sizeof("new Object({ 'state' : 'starting' })\r\n") - 1);
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, "reportuploads returning starting");
    } else if (err_status >= NGX_HTTP_SPECIAL_RESPONSE) {
        b->last = ngx_cpymem(b->last, "new Object({ 'state' : 'error', 'status' : ",
                             sizeof("new Object({ 'state' : 'error', 'status' : ") - 1);
        b->last = ngx_sprintf(b->last, "%ui })\r\n", err_status );
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                        "reportuploads returning error condition: %ui", err_status);
    } else if (done) {
        b->last = ngx_cpymem(b->last, "new Object({ 'state' : 'done' })\r\n",
                             sizeof("new Object({ 'state' : 'done' })\r\n") - 1);
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                        "reportuploads returning done");
    } else if (length == 0 && rest == 0) {
        b->last = ngx_cpymem(b->last, "new Object({ 'state' : 'starting' })\r\n",
                            sizeof("new Object({ 'state' : 'starting' })\r\n") - 1);
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0, "reportuploads returning starting");
    } else {
        b->last = ngx_cpymem(b->last, "new Object({ 'state' : 'uploading', 'received' : ",
                             sizeof("new Object({ 'state' : 'uploading', 'received' : ") - 1);

        b->last = ngx_sprintf(b->last, "%uO, 'size' : %uO })\r\n", (length - rest), length);

        ngx_log_debug2(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                        "reportuploads returning %uO / %uO",(length - rest), length );
    }


    r->headers_out.status = NGX_HTTP_OK;
    r->headers_out.content_length_n = b->last - b->pos;

    b->last_buf = 1;
    rc = ngx_http_send_header(r);

    if (rc == NGX_ERROR || rc > NGX_OK || r->header_only) {
        return rc;
    }

    return ngx_http_output_filter(r, &out);
}

extern void ngx_http_upload_progress_finish(ngx_http_request_t *r, ngx_http_uploadprogress_conf_t *track_info) {
	ngx_http_uploadprogress_ctx_t *ctx;
	ngx_slab_pool_t		*shpool;
	ngx_http_uploadprogress_node_t *node;
	ngx_str_t 		*id;
	
	
		
	id = get_tracking_id(r);
    if (id == NULL) {
        ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                       "trackuploads no id found in POST upload req");
        return;
    }
	
	ngx_log_error(NGX_LOG_INFO, r->connection->log, 0, "upload: try to mark node as 'done'");
		
	ctx = track_info->shm_zone->data;
	shpool = (ngx_slab_pool_t *) track_info->shm_zone->shm.addr;
			
	ngx_shmtx_lock(&shpool->mutex);
	
	node = find_node(id, ctx, ngx_cycle->log);
	if(node != NULL) {
		node->done = 1;               /* mark the original request as done */
		node->timeout = ngx_time() + track_info->timeout;      /* keep tracking for 60s */

		if (r != NULL ) {
			ngx_uint_t rc = r->err_status ? r->err_status : r->headers_out.status;
			if (rc >= NGX_HTTP_SPECIAL_RESPONSE) {
				node->err_status = rc;
			}
		}
	}
	ngx_shmtx_unlock(&shpool->mutex);
	
	ngx_log_error(NGX_LOG_INFO, r->connection->log, 0, "upload: done mark node as 'done'");
	ngx_free(id);

}

extern void change_progress(ngx_http_uploadprogress_conf_t *track_info, ngx_http_request_t *r) {
	ngx_shm_zone_t                  *shm_zone;
	ngx_http_uploadprogress_ctx_t	*ctx;
	ngx_http_uploadprogress_node_t 	*up;
	ngx_slab_pool_t                	*shpool;
	ngx_str_t *id;

	id = get_tracking_id(r);
    if (id == NULL) {
        ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                       "trackuploads no id found in POST upload req");
        return;
    }
	
	shm_zone = track_info->shm_zone;
	ctx = shm_zone->data;
	
	/* get the original connection of the upload */
	shpool = (ngx_slab_pool_t *) shm_zone->shm.addr;

	ngx_shmtx_lock(&shpool->mutex);
	up = find_node(id, ctx, ngx_cycle->log);
	if(up != NULL) {	
		up->rest = r->request_body->rest;
		up->length = r->headers_in.content_length_n;
		ngx_log_error(NGX_LOG_DEBUG	, r->connection->log, 0, "change_progress: %V %d / %d", id, up->rest, up->length);
	}
	ngx_shmtx_unlock(&shpool->mutex);
	ngx_free(id);
}
